==> divine-4.0.12.tar.gz: divine-4.0.12/divine/vm/dbg-util.hpp <==
// -*- mode: C++; indent-tabs-mode: nil; c-basic-offset: 4 -*-

/*
 * (c) 2017 Petr Ročkai <code@fixp.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once
#include <divine/vm/dbg-node.hpp>

namespace divine::vm::dbg
{

using DNSet = std::set< DNKey >;

template< typename DN >
void backtrace( DN dn, DNSet &visited, int &stacks, int maxdepth )
{
    if ( visited.count( dn.sortkey() ) || dn.address().type() != PointerType::Heap )
        return;
    visited.insert( dn.sortkey() );

    if ( !maxdepth )
        return;

    if ( dn.kind() == DNKind::Frame )
    {
        bool itemized = false;
        dn.attributes( [&]( std::string k, std::string v )
                       {
                           if ( k != "pc" && k != "address" && k != "location" && k != "symbol" )
                               return;
                           if ( itemized )
                               std::cout << "      ";
                           else
                               std::cout << "    - ", itemized = true;
                           std::cout << k << ": " << v << std::endl;
                       } );
        std::cout << std::endl;
    }

    auto follow =
        [&]( std::string k, auto rel )
        {
            if ( rel.kind() == DNKind::Frame && k != "caller" &&
                 rel.address().type() == PointerType::Heap &&
                 !visited.count( rel.sortkey() ) && maxdepth > 1 )
                std::cout << "  backtrace " << ++stacks << ":" << std::endl;
            backtrace( rel, visited, stacks, k == "caller" ? maxdepth - 1 : maxdepth );
        };

    dn.related( follow, false );
    dn.components( follow );
}

}

==> divine-4.0.19+2017.11.16.tar.gz: divine-4.0.19+2017.11.16/divine/vm/dbg-util.hpp <==
// -*- mode: C++; indent-tabs-mode: nil; c-basic-offset: 4 -*-

/*
 * (c) 2017 Petr Ročkai <code@fixp.eu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#pragma once
#include <divine/vm/dbg-node.hpp>

namespace divine::vm::dbg
{

using DNSet = std::set< DNKey >;

template< typename DN >
void backtrace( std::ostream &ostr, DN dn, DNSet &visited, int &stacks, int maxdepth )
{
    if ( visited.count( dn.sortkey() ) || dn.address().type() != PointerType::Heap )
        return;
    visited.insert( dn.sortkey() );

    if ( !maxdepth )
        return;

    if ( dn.kind() == DNKind::Frame )
    {
        bool itemized = false;
        dn.attributes( [&]( std::string k, std::string v )
                       {
                           if ( k != "pc" && k != "address" && k != "location" && k != "symbol" )
                               return;
                           if ( itemized )
                               ostr << "      ";
                           else
                               ostr << "    - ", itemized = true;
                           ostr << k << ": " << v << std::endl;
                       } );
        ostr << std::endl;
    }

    auto follow =
        [&]( std::string k, auto rel )
        {
            if ( rel.kind() == DNKind::Frame && k != "caller" &&
                 rel.address().type() == PointerType::Heap &&
                 !visited.count( rel.sortkey() ) && maxdepth > 1 )
                ostr << "  backtrace " << ++stacks << ":" << std::endl;
            backtrace( ostr, rel, visited, stacks, k == "caller" ? maxdepth - 1 : maxdepth );
        };

    dn.related( follow, false );
    dn.components( follow );
}

}
